package net.knarcraft.minecraftserverlauncher.userinterface;

import net.knarcraft.minecraftserverlauncher.Main;
import net.knarcraft.minecraftserverlauncher.profile.Collection;
import net.knarcraft.minecraftserverlauncher.profile.Controller;
import net.knarcraft.minecraftserverlauncher.profile.Profile;
import net.knarcraft.minecraftserverlauncher.server.Server;
import net.knarcraft.minecraftserverlauncher.utility.CommonFunctions;

import javax.imageio.ImageIO;
import javax.naming.ConfigurationException;
import javax.swing.*;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.*;
import java.util.*;
import java.util.concurrent.Executors;

import static java.awt.Frame.NORMAL;
import static javax.swing.WindowConstants.DO_NOTHING_ON_CLOSE;
import static net.knarcraft.minecraftserverlauncher.utility.CommonFunctions.getResourceAsScanner;
import static net.knarcraft.minecraftserverlauncher.utility.CommonFunctions.getResourceAsStream;

/**
 * Generates a ServerLauncherGUI.
 *
 * @author Kristian Knarvik <kristian.knarvik@knett.no>
 * @version 1.0.0
 * @since 1.0.0
 */
public class ServerLauncherGUI extends MessageHandler implements ActionListener, GUI {

    private final JLabel lblStatuslabel = new JLabel("Servers are stopped");
    private final ArrayList<String> globalPlayers;
    private Controller controller;
    private Map<String, String> textStrings;

    private JFrame frame;
    private JTabbedPane tabbedPane;
    private JTabbedPane serversPane;
    /*******
    * Menu
    ********/
    //Options
    private JCheckBoxMenuItem runInBackgroundCheckBoxMenuItem;
    private JCheckBoxMenuItem delayStartupCheckBoxMenuItem;
    private JCheckBoxMenuItem downloadJarsCheckBoxMenuItem;
    //Help
    private JMenuItem errorsMenuItem;
    private JMenuItem setupMenuItem;
    private JMenuItem manualUpdateMenuItem;
    //Info/options
    private JMenuItem runInBackgroundMenuItem;
    private JMenuItem delayStartupMenuItem;
    private JMenuItem downloadJarsMenuItem;
    //Info/about
    private JMenuItem aboutMenuItem;
    private JMenuItem storyMenuItem;
    //Basic controls
    private JButton startServerButton;
    private JButton stopServerButton;
    private JButton addServerButton;
    private JButton backupButton;
    private JButton addProfileButton;
    private JButton deleteProfileButton;
    private JComboBox<String> profiles;
    //Server controls
    private JComboBox<String> targetServerCombo;
    private JComboBox<String> targetPlayerCombo;
    private JButton kickButton;
    private JButton banButton;
    private JButton opButton;
    private JButton deopButton;
    private JButton customCommandButton;
    private JButton saveServerButton;
    private JButton reloadButton;
    private JButton showConsolesButton;
    private JTextField customCommandTextField;
    //Tray
    private SystemTray tray;
    private TrayIcon trayIcon;

    /**
     * Creates the application window
     */
    public ServerLauncherGUI() throws IOException {
        super(false);
        initialize(440, 170);
        loadMessages();
        this.controller = Main.getController();
        this.globalPlayers = new ArrayList<>();
    }

    /**
     * Creates the application window with a preferred width and height
     *
     * @param width  <p>The preferred width</p>
     * @param height <p>The preferred height</p>
     */
    public ServerLauncherGUI(int width, int height) throws IOException {
        super(false);
        initialize(width, height);
        loadMessages();
        this.controller = Main.getController();
        this.globalPlayers = new ArrayList<>();
    }

    /**
     * Gets the pane used for server configurations
     *
     * @return <p>The pane used for server configurations</p>
     */
    public JTabbedPane getPane() {
        return this.serversPane;
    }

    @Override
    public void setStatus(String text) {
        this.lblStatuslabel.setText(text);
        try (PrintWriter file = new PrintWriter(new FileWriter(Main.getApplicationWorkDirectory() + File.separator + "latestrun.log", true))) {
            file.println(text);
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    @Override
    public File askForDirectory(String prompt) {
        JFileChooser chooser = new JFileChooser();
        chooser.setCurrentDirectory(new java.io.File("."));
        chooser.setDialogTitle(prompt);
        chooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
        chooser.setAcceptAllFileFilterUsed(false);

        if (chooser.showOpenDialog(null) != JFileChooser.APPROVE_OPTION) {
            setStatus("Choice aborted by user");
            return null;
        }
        return chooser.getSelectedFile();
    }

    /**
     * Adds a player to the global player list, and updates the players combo
     *
     * @param name <p>The name of the player to add</p>
     */
    public void addPlayer(String name) {
        this.globalPlayers.add(name);
        this.updatePlayers();
    }

    /**
     * Removes a player from the global list of players
     *
     * @param name <p>The name of the player to remove</p>
     */
    public void removePlayer(String name) {
        globalPlayers.removeIf(playerName -> playerName.equals(name));
        this.updatePlayers();
    }

    /**
     * Updates the profiles combo
     */
    public void updateProfiles() {
        String selectedProfile = Main.getController().getCurrentProfile().getName();
        this.profiles.removeAllItems();
        for (String profile : Main.getController().getProfileNames()) {
            this.profiles.addItem(profile);
        }
        this.profiles.setSelectedItem(selectedProfile);
    }

    /**
     * Gets the size of the main JFrame
     *
     * @return <p>The Dimension of the main JFrame</p>
     */
    public Dimension getSize() {
        return frame.getContentPane().getSize();
    }

    /**
     * Updates ServerLauncherGUI according to current profile settings
     */
    public void update() {
        Controller controller = Main.getController();
        serversPane.removeAll();
        runInBackgroundCheckBoxMenuItem.setState(controller.getRunInBackground());
        delayStartupCheckBoxMenuItem.setState(controller.getDelayStartup() > 0);
        downloadJarsCheckBoxMenuItem.setState(controller.getDownloadAllJars());
        this.targetServerCombo.removeAllItems();
        this.targetServerCombo.addItem("All");
        for (Collection collection : controller.getCurrentProfile().getCollections()) {
            serversPane.addTab(collection.getName(), collection.getServerTab().getPanel());
            this.targetServerCombo.addItem(collection.getName());
        }
    }

    /**
     * Creates the ServerLauncherGUI
     */
    private void initialize(int width, int height) throws IOException {
        try {
            UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
        } catch (ClassNotFoundException |
                UnsupportedLookAndFeelException |
                InstantiationException |
                IllegalAccessException e
                ) {
            e.printStackTrace();
        }

        frame = new JFrame("Minecraft server launcher");
        frame.setDefaultCloseOperation(DO_NOTHING_ON_CLOSE);
        frame.getContentPane().setPreferredSize(new Dimension(width, height));

        ImageIcon img = new ImageIcon(ImageIO.read(getResourceAsStream("GUIIcon.png")));

        frame.setIconImage(img.getImage());

        JMenuBar menuBar = new JMenuBar();
        frame.setJMenuBar(menuBar);

        JMenu mnOptions = new JMenu("Options");
        menuBar.add(mnOptions);

        runInBackgroundCheckBoxMenuItem = new JCheckBoxMenuItem("Run in background on exit");
        mnOptions.add(runInBackgroundCheckBoxMenuItem);
        runInBackgroundCheckBoxMenuItem.addActionListener(this);

        delayStartupCheckBoxMenuItem = new JCheckBoxMenuItem("Delay Startup");
        mnOptions.add(delayStartupCheckBoxMenuItem);
        delayStartupCheckBoxMenuItem.addActionListener(this);

        downloadJarsCheckBoxMenuItem = new JCheckBoxMenuItem("Download jars");
        mnOptions.add(downloadJarsCheckBoxMenuItem);
        downloadJarsCheckBoxMenuItem.addActionListener(this);

        JMenu mnHelp = new JMenu("Help");
        menuBar.add(mnHelp);

        errorsMenuItem = new JMenuItem("Errors");
        mnHelp.add(errorsMenuItem);
        errorsMenuItem.addActionListener(this);

        setupMenuItem = new JMenuItem("Setup");
        mnHelp.add(setupMenuItem);
        setupMenuItem.addActionListener(this);

        manualUpdateMenuItem = new JMenuItem("Manual update");
        mnHelp.add(manualUpdateMenuItem);
        manualUpdateMenuItem.addActionListener(this);

        JMenu mnInfo = new JMenu("Info");
        menuBar.add(mnInfo);

        JMenu mnOptionsInfo = new JMenu("Options");
        mnInfo.add(mnOptionsInfo);

        runInBackgroundMenuItem = new JMenuItem("Run in background on exit");
        mnOptionsInfo.add(runInBackgroundMenuItem);
        runInBackgroundMenuItem.addActionListener(this);

        delayStartupMenuItem = new JMenuItem("Delay Startup");
        mnOptionsInfo.add(delayStartupMenuItem);
        delayStartupMenuItem.addActionListener(this);

        downloadJarsMenuItem = new JMenuItem("Download jars");
        mnOptionsInfo.add(downloadJarsMenuItem);
        downloadJarsMenuItem.addActionListener(this);

        JMenu mnAbout = new JMenu("About");
        mnInfo.add(mnAbout);

        aboutMenuItem = new JMenuItem("About");
        mnAbout.add(aboutMenuItem);
        aboutMenuItem.addActionListener(this);

        storyMenuItem = new JMenuItem("Story");
        mnAbout.add(storyMenuItem);
        storyMenuItem.addActionListener(this);

        tabbedPane = new JTabbedPane(JTabbedPane.TOP);
        frame.getContentPane().add(tabbedPane);

        JPanel panelBasic = new JPanel();
        tabbedPane.addTab("Control panel", null, panelBasic, null);
        SpringLayout sl_panel = new SpringLayout();
        panelBasic.setLayout(sl_panel);

        JLabel lblBasicControls = new JLabel("Basic controls");
        sl_panel.putConstraint(SpringLayout.NORTH, lblBasicControls, 10, SpringLayout.NORTH, panelBasic);
        panelBasic.add(lblBasicControls);

        startServerButton = new JButton("Start servers");
        sl_panel.putConstraint(SpringLayout.WEST, lblBasicControls, 0, SpringLayout.WEST, startServerButton);
        sl_panel.putConstraint(SpringLayout.NORTH, startServerButton, 6, SpringLayout.SOUTH, lblBasicControls);
        sl_panel.putConstraint(SpringLayout.WEST, startServerButton, 10, SpringLayout.WEST, panelBasic);
        panelBasic.add(startServerButton);
        startServerButton.addActionListener(this);

        stopServerButton = new JButton("Stop servers");
        sl_panel.putConstraint(SpringLayout.NORTH, stopServerButton, 0, SpringLayout.NORTH, startServerButton);
        sl_panel.putConstraint(SpringLayout.WEST, stopServerButton, 6, SpringLayout.EAST, startServerButton);
        panelBasic.add(stopServerButton);
        stopServerButton.addActionListener(this);

        JLabel lblProfile = new JLabel("Profile");
        sl_panel.putConstraint(SpringLayout.NORTH, lblProfile, 6, SpringLayout.SOUTH, startServerButton);
        sl_panel.putConstraint(SpringLayout.WEST, lblProfile, 10, SpringLayout.WEST, panelBasic);
        panelBasic.add(lblProfile);

        addProfileButton = new JButton("+");
        sl_panel.putConstraint(SpringLayout.NORTH, addProfileButton, 6, SpringLayout.SOUTH, lblProfile);
        sl_panel.putConstraint(SpringLayout.WEST, addProfileButton, 10, SpringLayout.WEST, panelBasic);
        panelBasic.add(addProfileButton);
        addProfileButton.addActionListener(this);

        deleteProfileButton = new JButton("-");
        sl_panel.putConstraint(SpringLayout.NORTH, deleteProfileButton, 0, SpringLayout.NORTH, addProfileButton);
        sl_panel.putConstraint(SpringLayout.WEST, deleteProfileButton, 6, SpringLayout.EAST, addProfileButton);
        panelBasic.add(deleteProfileButton);
        deleteProfileButton.addActionListener(this);

        profiles = new JComboBox<>();
        sl_panel.putConstraint(SpringLayout.NORTH, profiles, 0, SpringLayout.NORTH, addProfileButton);
        sl_panel.putConstraint(SpringLayout.WEST, profiles, 6, SpringLayout.EAST, deleteProfileButton);
        sl_panel.putConstraint(SpringLayout.EAST, profiles, 124, SpringLayout.EAST, deleteProfileButton);
        panelBasic.add(profiles);
        profiles.addActionListener(this);

        sl_panel.putConstraint(SpringLayout.NORTH, lblStatuslabel, 6, SpringLayout.SOUTH, addProfileButton);
        sl_panel.putConstraint(SpringLayout.SOUTH, lblStatuslabel, -10, SpringLayout.SOUTH, panelBasic);
        sl_panel.putConstraint(SpringLayout.WEST, lblStatuslabel, 10, SpringLayout.WEST, panelBasic);
        sl_panel.putConstraint(SpringLayout.EAST, lblStatuslabel, -10, SpringLayout.EAST, panelBasic);
        panelBasic.add(lblStatuslabel);

        addServerButton = new JButton("Add server");
        sl_panel.putConstraint(SpringLayout.NORTH, addServerButton, 0, SpringLayout.NORTH, startServerButton);
        sl_panel.putConstraint(SpringLayout.WEST, addServerButton, 6, SpringLayout.EAST, stopServerButton);
        panelBasic.add(addServerButton);
        addServerButton.addActionListener(this);

        backupButton = new JButton("Backup");
        sl_panel.putConstraint(SpringLayout.NORTH, backupButton, 0, SpringLayout.NORTH, startServerButton);
        sl_panel.putConstraint(SpringLayout.WEST, backupButton, 6, SpringLayout.EAST, addServerButton);
        panelBasic.add(backupButton);
        backupButton.addActionListener(this);

        JPanel controlServers = new JPanel();
        tabbedPane.addTab("Control servers", null, controlServers, null);
        SpringLayout sl_panel_1 = new SpringLayout();
        controlServers.setLayout(sl_panel_1);

        targetServerCombo = new JComboBox<>();
        sl_panel_1.putConstraint(SpringLayout.NORTH, targetServerCombo, 10, SpringLayout.NORTH, controlServers);
        controlServers.add(targetServerCombo);
        targetServerCombo.addActionListener(this);

        targetPlayerCombo = new JComboBox<>();
        sl_panel_1.putConstraint(SpringLayout.NORTH, targetPlayerCombo, 6, SpringLayout.SOUTH, targetServerCombo);
        targetPlayerCombo.setEditable(true);
        controlServers.add(targetPlayerCombo);

        kickButton = new JButton("Kick");
        sl_panel_1.putConstraint(SpringLayout.NORTH, kickButton, 10, SpringLayout.NORTH, controlServers);
        sl_panel_1.putConstraint(SpringLayout.WEST, kickButton, 6, SpringLayout.EAST, targetServerCombo);
        sl_panel_1.putConstraint(SpringLayout.EAST, kickButton, 104, SpringLayout.WEST, kickButton);
        sl_panel_1.putConstraint(SpringLayout.SOUTH, targetServerCombo, 0, SpringLayout.SOUTH, kickButton);
        controlServers.add(kickButton);
        kickButton.addActionListener(this);

        banButton = new JButton("Ban");
        sl_panel_1.putConstraint(SpringLayout.NORTH, banButton, 6, SpringLayout.SOUTH, kickButton);
        sl_panel_1.putConstraint(SpringLayout.WEST, banButton, 6, SpringLayout.EAST, targetPlayerCombo);
        sl_panel_1.putConstraint(SpringLayout.EAST, banButton, 104, SpringLayout.WEST, banButton);
        sl_panel_1.putConstraint(SpringLayout.SOUTH, targetPlayerCombo, 0, SpringLayout.SOUTH, banButton);
        controlServers.add(banButton);
        banButton.addActionListener(this);

        opButton = new JButton("OP");
        sl_panel_1.putConstraint(SpringLayout.NORTH, opButton, 10, SpringLayout.NORTH, controlServers);
        sl_panel_1.putConstraint(SpringLayout.WEST, opButton, 6, SpringLayout.EAST, kickButton);
        sl_panel_1.putConstraint(SpringLayout.EAST, opButton, -10, SpringLayout.EAST, controlServers);
        controlServers.add(opButton);
        opButton.addActionListener(this);

        deopButton = new JButton("DEOP");
        sl_panel_1.putConstraint(SpringLayout.WEST, deopButton, 6, SpringLayout.EAST, banButton);
        sl_panel_1.putConstraint(SpringLayout.NORTH, deopButton, 5, SpringLayout.SOUTH, opButton);
        sl_panel_1.putConstraint(SpringLayout.EAST, deopButton, -10, SpringLayout.EAST, controlServers);
        controlServers.add(deopButton);
        deopButton.addActionListener(this);

        JLabel lblTargetServer = new JLabel("Target server");
        sl_panel_1.putConstraint(SpringLayout.WEST, targetServerCombo, 6, SpringLayout.EAST, lblTargetServer);
        sl_panel_1.putConstraint(SpringLayout.EAST, targetServerCombo, 121, SpringLayout.EAST, lblTargetServer);
        sl_panel_1.putConstraint(SpringLayout.NORTH, lblTargetServer, 10, SpringLayout.NORTH, controlServers);
        sl_panel_1.putConstraint(SpringLayout.SOUTH, lblTargetServer, 0, SpringLayout.SOUTH, targetServerCombo);
        sl_panel_1.putConstraint(SpringLayout.WEST, lblTargetServer, 10, SpringLayout.WEST, controlServers);
        controlServers.add(lblTargetServer);

        JLabel lblTargetPlayer = new JLabel("Target player");
        sl_panel_1.putConstraint(SpringLayout.WEST, targetPlayerCombo, 7, SpringLayout.EAST, lblTargetPlayer);
        sl_panel_1.putConstraint(SpringLayout.EAST, targetPlayerCombo, 122, SpringLayout.EAST, lblTargetPlayer);
        sl_panel_1.putConstraint(SpringLayout.NORTH, lblTargetPlayer, 6, SpringLayout.SOUTH, lblTargetServer);
        sl_panel_1.putConstraint(SpringLayout.SOUTH, lblTargetPlayer, 0, SpringLayout.SOUTH, targetPlayerCombo);
        sl_panel_1.putConstraint(SpringLayout.WEST, lblTargetPlayer, 0, SpringLayout.WEST, lblTargetServer);
        controlServers.add(lblTargetPlayer);

        customCommandButton = new JButton("Custom command");
        sl_panel_1.putConstraint(SpringLayout.WEST, customCommandButton, 250, SpringLayout.WEST, controlServers);
        sl_panel_1.putConstraint(SpringLayout.EAST, customCommandButton, 0, SpringLayout.EAST, opButton);
        controlServers.add(customCommandButton);
        customCommandButton.addActionListener(this);
        frame.getRootPane().setDefaultButton(customCommandButton);

        customCommandTextField = new JTextField();
        sl_panel_1.putConstraint(SpringLayout.WEST, customCommandTextField, 10, SpringLayout.WEST, controlServers);
        sl_panel_1.putConstraint(SpringLayout.EAST, customCommandTextField, -6, SpringLayout.WEST, customCommandButton);
        sl_panel_1.putConstraint(SpringLayout.NORTH, customCommandButton, 0, SpringLayout.NORTH, customCommandTextField);
        sl_panel_1.putConstraint(SpringLayout.SOUTH, customCommandTextField, 0, SpringLayout.SOUTH, customCommandButton);
        controlServers.add(customCommandTextField);
        customCommandTextField.setColumns(10);

        saveServerButton = new JButton("Save server");
        sl_panel_1.putConstraint(SpringLayout.NORTH, customCommandTextField, 6, SpringLayout.SOUTH, saveServerButton);
        sl_panel_1.putConstraint(SpringLayout.NORTH, saveServerButton, 6, SpringLayout.SOUTH, banButton);
        sl_panel_1.putConstraint(SpringLayout.WEST, saveServerButton, 0, SpringLayout.WEST, kickButton);
        sl_panel_1.putConstraint(SpringLayout.EAST, saveServerButton, 104, SpringLayout.WEST, kickButton);
        sl_panel_1.putConstraint(SpringLayout.EAST, saveServerButton, 104, SpringLayout.WEST, kickButton);
        controlServers.add(saveServerButton);
        saveServerButton.addActionListener(this);

        reloadButton = new JButton("Reload");
        sl_panel_1.putConstraint(SpringLayout.NORTH, reloadButton, 6, SpringLayout.SOUTH, deopButton);
        sl_panel_1.putConstraint(SpringLayout.WEST, reloadButton, 0, SpringLayout.WEST, deopButton);
        sl_panel_1.putConstraint(SpringLayout.EAST, reloadButton, 0, SpringLayout.EAST, opButton);
        controlServers.add(reloadButton);
        reloadButton.addActionListener(this);

        showConsolesButton = new JButton("View server consoles");
        sl_panel_1.putConstraint(SpringLayout.NORTH, showConsolesButton, 0, SpringLayout.NORTH, saveServerButton);
        sl_panel_1.putConstraint(SpringLayout.WEST, showConsolesButton, 0, SpringLayout.WEST, lblTargetServer);
        sl_panel_1.putConstraint(SpringLayout.EAST, showConsolesButton, 0, SpringLayout.EAST, targetServerCombo);
        controlServers.add(showConsolesButton);
        showConsolesButton.addActionListener(this);

        JPanel panel_2 = new JPanel();
        tabbedPane.addTab("Servers", null, panel_2, null);
        SpringLayout sl_panel_2 = new SpringLayout();
        panel_2.setLayout(sl_panel_2);

        JTabbedPane tabbedPane_1 = new JTabbedPane(JTabbedPane.TOP);
        sl_panel_2.putConstraint(SpringLayout.NORTH, tabbedPane_1, 0, SpringLayout.NORTH, panel_2);
        sl_panel_2.putConstraint(SpringLayout.WEST, tabbedPane_1, 0, SpringLayout.WEST, panel_2);
        sl_panel_2.putConstraint(SpringLayout.SOUTH, tabbedPane_1, 0, SpringLayout.SOUTH, panel_2);
        sl_panel_2.putConstraint(SpringLayout.EAST, tabbedPane_1, 0, SpringLayout.EAST, panel_2);
        panel_2.add(tabbedPane_1);

        this.serversPane = tabbedPane_1;
        tabbedPane_1.setTabLayoutPolicy(JTabbedPane.SCROLL_TAB_LAYOUT);

        frame.validate();
        frame.pack();
        frame.setVisible(true);
        tray();
        updateRunning(false);
    }

    /**
     * Prepares the system tray if available
     */
    private void tray() {
        if (SystemTray.isSupported()) {
            tray = SystemTray.getSystemTray();
            InputStream imageStream = CommonFunctions.getResourceAsStream("GUIIcon.png");
            Image trayImage;
            try {
                trayImage = ImageIO.read(imageStream);
            } catch (IOException e) {
                trayImage = Toolkit.getDefaultToolkit().getImage("resources/GUIIcon.png");
                e.printStackTrace();
            }
            PopupMenu popup = new PopupMenu();
            trayIcon = new TrayIcon(trayImage, "Minecraft Server Launcher", popup);
            trayIcon.setImageAutoSize(true);
            ActionListener exitListener = e -> {
                stop();
                controller.saveState();
                System.exit(0);
            };

            MenuItem restoreItem = new MenuItem("Restore");
            popup.add(restoreItem);
            restoreItem.addActionListener(e -> {
                frame.setExtendedState(NORMAL);
                tray.remove(trayIcon);
                frame.setVisible(true);
            });
            MenuItem exitItem = new MenuItem("Exit");
            exitItem.addActionListener(exitListener);
            popup.add(exitItem);
            frame.addWindowStateListener(e -> {
                if (e.getNewState() == NORMAL) {
                    tray.remove(trayIcon);
                    frame.setVisible(true);
                }
            });

            frame.addWindowListener(new WindowAdapter() {
                @Override
                public void windowClosing(WindowEvent e) {
                    if (controller.getRunInBackground() && SystemTray.isSupported()) {
                        try {
                            tray.add(trayIcon);
                            frame.setVisible(false);
                        } catch (AWTException e1) {
                            e1.printStackTrace();
                        }
                    } else {
                        stop();
                        controller.saveState();
                        System.exit(0);
                    }
                }
            });

            trayIcon.addMouseListener(new MouseAdapter() {
                @Override
                public void mousePressed(MouseEvent e) {
                    if (e.getClickCount() >= 1 && e.getButton() == MouseEvent.BUTTON1) {
                        frame.setExtendedState(NORMAL);
                        tray.remove(trayIcon);
                        frame.setVisible(true);
                    }
                }
            });
        } else {
            frame.addWindowListener(new WindowAdapter() {
                @Override
                public void windowClosing(WindowEvent e) {
                    controller.saveState();
                    stop();
                    System.exit(0);
                }
            });
        }
    }

    /**
     * Hides the gui to the tray
     */
    public void hide() {
        frame.setVisible(false);
        try {
            tray.add(trayIcon);
        } catch (AWTException e) {
            e.printStackTrace();
        }
    }

    @Override
    public void actionPerformed(ActionEvent e) {
        String selectedServerValue = null, selectedPlayerValue = null;
        Object selectedServer = targetServerCombo.getSelectedItem();
        if (selectedServer != null) {
            selectedServerValue = selectedServer.toString();
        }
        Object selectedPlayer = targetPlayerCombo.getSelectedItem();
        if (selectedPlayer != null) {
            selectedPlayerValue = selectedPlayer.toString();
        }
        if (e.getSource() == runInBackgroundCheckBoxMenuItem) {
            background();
        } else if (e.getSource() == delayStartupCheckBoxMenuItem) {
            delay();
        } else if (e.getSource() == downloadJarsCheckBoxMenuItem) {
            downloadJars();
        } else if (e.getSource() == errorsMenuItem) {
            CommonFunctions.goToURL(textStrings.get("infoURL"));
        } else if (e.getSource() == setupMenuItem) {
            showMessage("Setup", textStrings.get("setupText"));
        } else if (e.getSource() == manualUpdateMenuItem) {
            CommonFunctions.goToURL(textStrings.get("manualUpdateURL"));
        } else if (e.getSource() == runInBackgroundMenuItem) {
            showMessage("Run in background", textStrings.get("runInBackgroundText"));
        } else if (e.getSource() == delayStartupMenuItem) {
            showMessage("Delay startup", textStrings.get("delayStartupText"));
        } else if (e.getSource() == downloadJarsMenuItem) {
            showMessage("Download jars", textStrings.get("downloadJarsText"));
        } else if (e.getSource() == aboutMenuItem) {
            showMessage("About", textStrings.get("aboutText"));
        } else if (e.getSource() == storyMenuItem) {
            CommonFunctions.goToURL(textStrings.get("storyURL"));
        } else if (e.getSource() == startServerButton) {
            controller.saveState();
            Executors.newSingleThreadExecutor().execute(Server::startServers);
        } else if (e.getSource() == stopServerButton) {
            stop();
        } else if (e.getSource() == addServerButton) {
            String serverName = JOptionPane.showInputDialog("Name of server: ");
            try {
                controller.getCurrentProfile().addCollection(serverName);
            } catch (ConfigurationException e1) {
                e1.printStackTrace();
            }
            this.update();
            controller.getCurrentProfile().updateConsoles();
        } else if (e.getSource() == backupButton) {
            CommonFunctions.backup(this);
        } else if (e.getSource() == addProfileButton) {
            controller.addProfile(JOptionPane.showInputDialog("Profile name: "));
            updateProfiles();
        } else if (e.getSource() == deleteProfileButton) {
            Object selected = profiles.getSelectedItem();
            if (selected != null) {
                controller.removeProfile(selected.toString());
                updateProfiles();
            }
        } else if (e.getSource() == profiles) {
            changeProfile();
        } else if (e.getSource() == kickButton) {
            if (selectedServerValue != null && selectedPlayerValue != null) {
                controller.getCurrentProfile().sendCommand(selectedServerValue, "kick " + selectedPlayerValue);
            }
        } else if (e.getSource() == banButton) {
            if (selectedServerValue != null && selectedPlayerValue != null) {
                controller.getCurrentProfile().sendCommand(selectedServerValue, "ban " + selectedPlayerValue);
            }
        } else if (e.getSource() == opButton) {
            if (selectedServerValue != null && selectedPlayerValue != null) {
                controller.getCurrentProfile().sendCommand(selectedServerValue, "op " + selectedPlayerValue);
            }
        } else if (e.getSource() == deopButton) {
            if (selectedServerValue != null && selectedPlayerValue != null) {
                controller.getCurrentProfile().sendCommand(selectedServerValue, "deop " + selectedPlayerValue);
            }
        } else if (e.getSource() == customCommandButton) {
            if (selectedServerValue != null) {
                controller.getCurrentProfile().sendCommand(selectedServerValue, customCommandTextField.getText());
                customCommandTextField.setText("");
            }
        } else if (e.getSource() == saveServerButton) {
            if (selectedServerValue != null) {
                controller.getCurrentProfile().sendCommand(selectedServerValue, "save-all");
            }
        } else if (e.getSource() == reloadButton) {
            if (selectedServerValue != null) {
                controller.getCurrentProfile().sendCommand(selectedServerValue, "reload");
            }
        } else if (e.getSource() == showConsolesButton) {
            ServerConsoles.setAsVisible();
        } else if (e.getSource() == targetServerCombo) {
            updatePlayers();
        }
    }

    /**
     * Updates the ServerLauncherGUI components to block a user from doing illegal actions
     *
     * @param running <p>Whether the servers are currently running</p>
     */
    public void updateRunning(boolean running) {
        boolean stopped = !running; //Most gui is only enabled when the server is stopped rather than running.
        profiles.setEnabled(stopped);
        addProfileButton.setEnabled(stopped);
        deleteProfileButton.setEnabled(stopped);
        startServerButton.setEnabled(stopped);
        addServerButton.setEnabled(stopped);
        tabbedPane.setEnabledAt(2, stopped);
        stopServerButton.setEnabled(running);
    }

    /**
     * Saves the previous profile and loads data from the new profile
     */
    private void changeProfile() {
        controller.saveState();
        Object current = this.profiles.getSelectedItem();
        if (current != null) {
            controller.setCurrentProfile(current.toString());
        }
        this.update();
        controller.getCurrentProfile().updateConsoles();
    }

    /**
     * Stops all servers
     */
    private void stop() {
        try {
            setStatus("Servers are stopping...");
            Server.stop();
        } catch (IOException e1) {
            showError("Could not stop server.");
            e1.printStackTrace();
        }
    }

    /**
     * Asks the user for a delay if checked, and sets the value to the current profile
     */
    private void delay() {
        Object selected = profiles.getSelectedItem();
        if (selected != null) {
            Profile profile = controller.getProfileByName(selected.toString());
            if (delayStartupCheckBoxMenuItem.isSelected()) {
                String response = JOptionPane.showInputDialog("Seconds to delay: ");
                if (response == null) {
                    delayStartupCheckBoxMenuItem.setState(false);
                    return;
                }
                int parsed = Integer.parseInt(response);
                Objects.requireNonNull(profile).setDelayStartup(parsed);
            } else {
                Objects.requireNonNull(profile).setDelayStartup(0);
            }
        } else {
            showError("No profile selected");
        }
    }

    /**
     * Saves the runInBackground setting to the current profile
     */
    private void background() {
        Object selected = profiles.getSelectedItem();
        if (selected != null) {
            Profile profile = controller.getProfileByName(selected.toString());
            Objects.requireNonNull(profile).setRunInBackground(runInBackgroundCheckBoxMenuItem.isSelected());
        } else {
            showError("No profile selected");
        }
    }

    /**
     * Saves the downloadJars setting to the current profile
     */
    private void downloadJars() {
        Object selected = profiles.getSelectedItem();
        if (selected != null) {
            controller.setDownloadAllJars(downloadJarsCheckBoxMenuItem.isSelected());
        } else {
            showError("No profile selected");
        }
    }

    /**
     * Updates the list of players currently online on the selected server
     */
    private void updatePlayers() {
        String selectedServerValue;
        Object selectedServer = targetServerCombo.getSelectedItem();
        if (selectedServer != null) {
            targetPlayerCombo.removeAllItems();
            selectedServerValue = selectedServer.toString();
            if (selectedServerValue.equals("All")) {
                for (String player : this.globalPlayers) {
                    targetPlayerCombo.addItem(player);
                }
            } else {
                for (String player : controller.getCurrentProfile().getCollection(selectedServerValue).getServer().getPlayers()) {
                    targetPlayerCombo.addItem(player);
                }
            }
        }
    }

    /**
     * Loads popup messages from a text file
     */
    private void loadMessages() throws FileNotFoundException {
        textStrings = new HashMap<>();
        Scanner file = getResourceAsScanner("menumsg.csv");
        while (file.hasNextLine()) {
            String nextLine = file.nextLine();
            String[] line = nextLine.split("=");
            String content = line[1].replaceAll("_BREAK_", System.getProperty("line.separator"));
            textStrings.put(line[0], content);
        }
    }
}
